import numpy as np
from PIL import Image
from PIL.ImageColor import getrgb
from PIL.Image import Image as IMG
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from .types import ColorType, SizeType, XYType


class ColorStop:
    def __init__(self, stop: float, color: "ColorType"):
        self.stop = stop
        """介于 0.0 与 1.0 之间的值，表示渐变中开始与结束之间的位置"""
        if isinstance(color, str):
            color = getrgb(color)
        if len(color) == 3:
            color = (color[0], color[1], color[2], 255)
        self.color = color
        """在 stop 位置显示的颜色值"""

    def __lt__(self, other: "ColorStop"):
        return self.stop < other.stop


class Gradient:
    def __init__(self, color_stops: List[ColorStop] = []):
        self.color_stops = color_stops
        self.color_stops.sort()

    def add_color_stop(self, stop: float, color: "ColorType"):
        self.color_stops.append(ColorStop(stop, color))
        self.color_stops.sort()

    def create_image(self, size: "SizeType") -> IMG:
        raise NotImplementedError


class LinearGradient(Gradient):
    def __init__(self, xy: "XYType", color_stops: List[ColorStop] = []):
        self.xy = xy
        self.x0 = xy[0]
        """渐变开始点的 x 坐标"""
        self.y0 = xy[1]
        """渐变开始点的 y 坐标"""
        self.x1 = xy[2]
        """渐变结束点的 x 坐标"""
        self.y1 = xy[3]
        """渐变结束点的 y 坐标"""
        super().__init__(color_stops)

    def create_image(self, size: "SizeType") -> IMG:
        w = size[0]
        h = size[1]
        x0 = self.x0
        y0 = self.y0
        x1 = self.x1
        y1 = self.y1
        x01 = x1 - x0
        y01 = y1 - y0
        d = x01**2 + y01**2
        arr = np.zeros([h, w, 4], np.uint8)
        if len(self.color_stops) == 1:
            arr[:, :, :] = self.color_stops[0].color
        elif self.color_stops:
            for x in range(w):
                for y in range(h):
                    x0p = x - x0
                    y0p = y - y0
                    ratio = (x0p * x01 + y0p * y01) / d
                    color = (0, 0, 0, 0)
                    if ratio < 0:
                        color = self.color_stops[0].color
                    if ratio >= 1:
                        color = self.color_stops[-1].color
                    for i in range(len(self.color_stops) - 1):
                        color_stop0 = self.color_stops[i]
                        color_stop1 = self.color_stops[i + 1]
                        stop0 = color_stop0.stop
                        stop1 = color_stop1.stop
                        color0 = color_stop0.color
                        color1 = color_stop1.color
                        if stop0 <= ratio < stop1:
                            e = (ratio - stop0) / (stop1 - stop0)
                            color = [
                                round((1 - e) * color0[j] + e * color1[j])
                                for j in range(4)
                            ]
                    arr[y, x, :] = color
        return Image.fromarray(arr)


if __name__ == "__main__":
    img = LinearGradient(
        (0, 0, 255, 255),
        [
            ColorStop(0, "red"),
            ColorStop(0.1, "orange"),
            ColorStop(0.25, "yellow"),
            ColorStop(0.4, "green"),
            ColorStop(0.6, "blue"),
            ColorStop(0.8, "cyan"),
            ColorStop(1, "purple"),
        ],
    ).create_image((255, 255))
    img.save("test.png")



import cv2
import numpy as np
from io import BytesIO
from pathlib import Path
from PIL import Image, ImageDraw
from PIL.ImageColor import getrgb
from PIL.Image import Image as IMG
from PIL.ImageFilter import Filter
from PIL.ImageDraw import ImageDraw as Draw
from typing import List, Optional, Type, Union

from .types import *
from .text2image import Text2Image
from .gradient import ColorStop, Gradient


class BuildImage:
    def __init__(self, image: IMG):
        self.image = image

    @property
    def width(self) -> int:
        return self.image.width

    @property
    def height(self) -> int:
        return self.image.height

    @property
    def size(self) -> SizeType:
        return self.image.size

    @property
    def mode(self) -> ModeType:
        return self.image.mode  # type: ignore

    @property
    def draw(self) -> Draw:
        return ImageDraw.Draw(self.image)

    @classmethod
    def new(
        cls, mode: ModeType, size: SizeType, color: Optional[ColorType] = None
    ) -> "BuildImage":
        return cls(Image.new(mode, size, color))  # type: ignore

    @classmethod
    def open(cls, file: Union[str, bytes, BytesIO, Path]) -> "BuildImage":
        return cls(Image.open(file))

    def copy(self) -> "BuildImage":
        return BuildImage(self.image.copy())

    def resize(
        self,
        size: SizeType,
        resample: ResampleType = Image.ANTIALIAS,
        keep_ratio: bool = False,
        inside: bool = False,
        direction: DirectionType = "center",
        bg_color: Optional[ColorType] = None,
        **kwargs
    ) -> "BuildImage":
        """
        调整图片尺寸

        :参数:
          * ``size``: 期望图片大小
          * ``keep_ratio``: 是否保持长宽比，默认为 `False`
          * ``inside``: `keep_ratio` 为 `True` 时，
                        若 `inside` 为 `True`，则调整图片大小至包含于期望尺寸，不足部分设为指定颜色；
                        若 `inside` 为 `False`，则调整图片大小至包含期望尺寸，超出部分裁剪
          * ``direction``: 调整图片大小时图片的方位；默认为居中
          * ``bg_color``: 不足部分设置的颜色
        """
        width, height = size
        if keep_ratio:
            if inside:
                ratio = min(width / self.width, height / self.height)
            else:
                ratio = max(width / self.width, height / self.height)
            width = int(self.width * ratio)
            height = int(self.height * ratio)

        image = BuildImage(
            self.image.resize((width, height), resample=resample, **kwargs)
        )

        if keep_ratio:
            image = image.resize_canvas(size, direction, bg_color, **kwargs)
        return image

    def resize_canvas(
        self,
        size: SizeType,
        direction: DirectionType = "center",
        bg_color: Optional[ColorType] = None,
    ) -> "BuildImage":
        """
        调整“画布”大小，超出部分裁剪，不足部分设为指定颜色

        :参数:
          * ``size``: 期望图片大小
          * ``direction``: 调整图片大小时图片的方位；默认为居中
          * ``bg_color``: 不足部分设置的颜色
        """
        w, h = size
        x = int((w - self.width) / 2)
        y = int((h - self.height) / 2)
        if direction in ["north", "northwest", "northeast"]:
            y = 0
        elif direction in ["south", "southwest", "southeast"]:
            y = h - self.height
        if direction in ["west", "northwest", "southwest"]:
            x = 0
        elif direction in ["east", "northeast", "southeast"]:
            x = w - self.width
        image = BuildImage.new(self.mode, size, bg_color)
        image.paste(self.image, (x, y))
        return image

    def resize_width(self, width: int, **kwargs) -> "BuildImage":
        """调整图片宽度，不改变长宽比"""
        return self.resize((width, int(self.height * width / self.width)), **kwargs)

    def resize_height(self, height: int, **kwargs) -> "BuildImage":
        """调整图片高度，不改变长宽比"""
        return self.resize((int(self.width * height / self.height), height), **kwargs)

    def rotate(
        self,
        angle: float,
        resample: ResampleType = Image.BICUBIC,
        expand: bool = False,
        **kwargs
    ) -> "BuildImage":
        """旋转图片"""
        image = BuildImage(
            self.image.rotate(angle, resample=resample, expand=expand, **kwargs)
        )
        return image

    def square(self) -> "BuildImage":
        """将图片裁剪为方形"""
        length = min(self.width, self.height)
        return self.resize_canvas((length, length))

    def circle(self) -> "BuildImage":
        """将图片裁剪为圆形"""
        image = self.square().image.convert("RGBA")
        mask = Image.new("L", (self.width * 10, self.height * 10), 0)
        draw = ImageDraw.Draw(mask)
        draw.ellipse((0, 0, mask.width, mask.height), 255)
        mask = mask.resize(self.size, Image.ANTIALIAS)
        bg = Image.new("RGBA", self.size)
        return BuildImage(Image.composite(image, bg, mask))

    def circle_corner(self, r: float) -> "BuildImage":
        """将图片裁剪为圆角矩形"""
        image = self.image.convert("RGBA")
        mask = Image.new("L", (self.width * 10, self.height * 10), 0)
        draw = ImageDraw.Draw(mask)
        draw.rounded_rectangle((0, 0, mask.width, mask.height), r * 10, fill=255)
        mask = mask.resize(self.size, Image.ANTIALIAS)
        bg = Image.new("RGBA", self.size)
        return BuildImage(Image.composite(image, bg, mask))

    def crop(self, box: BoxType) -> "BuildImage":
        """裁剪图片"""
        return BuildImage(self.image.crop(box))

    def convert(self, mode: ModeType, **kwargs) -> "BuildImage":
        return BuildImage(self.image.convert(mode, **kwargs))

    def paste(
        self,
        img: Union[IMG, "BuildImage"],
        pos: PosTypeInt = (0, 0),
        alpha: bool = False,
        below: bool = False,
    ) -> "BuildImage":
        """
        粘贴图片

        :参数:
          * ``img``: 待粘贴的图片
          * ``pos``: 粘贴位置
          * ``alpha``: 图片背景是否为透明
          * ``below``: 是否粘贴到底层
        """
        if isinstance(img, BuildImage):
            img = img.image
        new_img = Image.new(self.mode, self.size) if below else self.image.copy()
        if alpha:
            img = img.convert("RGBA")
            new_img.paste(img, pos, mask=img)
        else:
            new_img.paste(img, pos)
        if below:
            new_img.paste(self.image, mask=self.image if self.mode == "RGBA" else None)
        self.image = new_img
        return self

    def alpha_composite(
        self,
        img: Union[IMG, "BuildImage"],
        dest: PosTypeInt = (0, 0),
        source: Union[PosTypeInt, BoxType] = (0, 0),
    ) -> "BuildImage":
        if isinstance(img, BuildImage):
            img = img.image
        return BuildImage(self.image.alpha_composite(img, dest=dest, source=source))  # type: ignore

    def filter(self, filter: Union[Filter, Type[Filter]]) -> "BuildImage":
        """滤波"""
        return BuildImage(self.image.filter(filter))

    def transpose(self, method: TransposeType) -> "BuildImage":
        """变换"""
        return BuildImage(self.image.transpose(method))

    def perspective(self, points: PointsTYpe) -> "BuildImage":
        """
        透视变换

        :参数:
          * ``points``: 变换后点的位置，顺序依次为：左上->右上->右下->左下
        """

        def find_coeffs(pa: PointsTYpe, pb: PointsTYpe):
            matrix = []
            for p1, p2 in zip(pa, pb):
                matrix.append(
                    [p1[0], p1[1], 1, 0, 0, 0, -p2[0] * p1[0], -p2[0] * p1[1]]
                )
                matrix.append(
                    [0, 0, 0, p1[0], p1[1], 1, -p2[1] * p1[0], -p2[1] * p1[1]]
                )
            A = np.matrix(matrix, dtype=np.float32)
            B = np.array(pb).reshape(8)
            res = np.dot(np.linalg.inv(A.T * A) * A.T, B)
            return np.array(res).reshape(8)

        img_w, img_h = self.size
        points_w = [p[0] for p in points]
        points_h = [p[1] for p in points]
        new_w = int(max(points_w) - min(points_w))
        new_h = int(max(points_h) - min(points_h))
        p = ((0, 0), (img_w, 0), (img_w, img_h), (0, img_h))
        coeffs = find_coeffs(points, p)
        return BuildImage(
            self.image.transform(
                (new_w, new_h), Image.PERSPECTIVE, coeffs, Image.BICUBIC
            )
        )

    def gradient_color(self, gradient: Gradient) -> "BuildImage":
        """
        渐变色

        :参数:
          * ``gradient``: 渐变对象
        """
        return BuildImage(gradient.create_image(self.size))

    def motion_blur(self, angle: float = 0, degree: int = 0) -> "BuildImage":
        """
        运动模糊

        :参数:
          * ``angle``: 运动方向
          * ``degree``: 模糊程度
        """
        if degree == 0:
            return self.copy()
        matrix = cv2.getRotationMatrix2D((degree / 2, degree / 2), angle + 45, 1)
        kernel = np.diag(np.ones(degree))
        kernel = cv2.warpAffine(kernel, matrix, (degree, degree)) / degree
        blurred = cv2.filter2D(np.asarray(self.image), -1, kernel)
        cv2.normalize(blurred, blurred, 0, 255, cv2.NORM_MINMAX)
        return BuildImage(Image.fromarray(np.array(blurred, dtype=np.uint8)))

    def distort(self, coefficients: DistortType) -> "BuildImage":
        """
        畸变

        :参数:
          * ``coefficients``: 畸变参数
        """
        res = cv2.undistort(
            np.asarray(self.image),
            np.array([[100, 0, self.width / 2], [0, 100, self.height / 2], [0, 0, 1]]),
            np.asarray(coefficients),
        )
        return BuildImage(Image.fromarray(np.array(res, dtype=np.uint8)))

    def color_mask(self, color: ColorType) -> "BuildImage":
        """
        颜色滤镜，改变图片色调

        :参数:
          * ``color``: 目标颜色
        """
        img = self.image.convert("RGB")
        w, h = img.size
        img_array = np.asarray(img)
        img_gray = cv2.cvtColor(img_array, cv2.COLOR_RGB2GRAY)
        img_hsl = cv2.cvtColor(img_array, cv2.COLOR_RGB2HLS)
        img_new = np.zeros((h, w, 3), np.uint8)

        if isinstance(color, str):
            color = getrgb(color)
        r = color[0]
        g = color[1]
        b = color[2]
        rgb_sum = sum(color)
        for i in range(h):
            for j in range(w):
                value = img_gray[i, j]
                new_color = [
                    int(value * r / rgb_sum),
                    int(value * g / rgb_sum),
                    int(value * b / rgb_sum),
                ]
                img_new[i, j] = new_color
        img_new_hsl = cv2.cvtColor(img_new, cv2.COLOR_RGB2HLS)
        result = np.dstack(
            (img_new_hsl[:, :, 0], img_hsl[:, :, 1], img_new_hsl[:, :, 2])
        )
        result = cv2.cvtColor(result, cv2.COLOR_HLS2RGB)
        return BuildImage(Image.fromarray(result))

    def draw_point(
        self, pos: PosTypeFloat, fill: Optional[ColorType] = None
    ) -> "BuildImage":
        """在图片上画点"""
        self.draw.point(pos, fill=fill)
        return self

    def draw_line(
        self,
        xy: XYType,
        fill: Optional[ColorType] = None,
        width: float = 1,
    ) -> "BuildImage":
        """在图片上画直线"""
        self.draw.line(xy, fill=fill, width=width)
        return self

    def draw_rectangle(
        self,
        xy: XYType,
        fill: Optional[ColorType] = None,
        outline: Optional[ColorType] = None,
        width: float = 1,
    ) -> "BuildImage":
        """在图片上画矩形"""
        self.draw.rectangle(xy, fill, outline, width)
        return self

    def draw_rounded_rectangle(
        self,
        xy: XYType,
        radius: int = 0,
        fill: Optional[ColorType] = None,
        outline: Optional[ColorType] = None,
        width: float = 1,
    ) -> "BuildImage":
        """在图片上画圆角矩形"""
        self.draw.rounded_rectangle(xy, radius, fill, outline, width)
        return self

    def draw_polygon(
        self,
        xy: List[PosTypeFloat],
        fill: Optional[ColorType] = None,
        outline: Optional[ColorType] = None,
        width: float = 1,
    ) -> "BuildImage":
        """在图片上画多边形"""
        self.draw.polygon(xy, fill, outline, width)
        return self

    def draw_arc(
        self,
        xy: XYType,
        start: float,
        end: float,
        fill: Optional[ColorType] = None,
        width: float = 1,
    ) -> "BuildImage":
        """在图片上画圆弧"""
        self.draw.arc(xy, start, end, fill, width)
        return self

    def draw_ellipse(
        self,
        xy: XYType,
        fill: Optional[ColorType] = None,
        outline: Optional[ColorType] = None,
        width: float = 1,
    ) -> "BuildImage":
        """在图片上画圆"""
        self.draw.ellipse(xy, fill, outline, width)
        return self

    def draw_text(
        self,
        xy: XYType,
        text: str,
        max_fontsize: int = 30,
        min_fontsize: int = 12,
        allow_wrap: bool = False,
        style: FontStyle = "normal",
        weight: FontWeight = "normal",
        fill: ColorType = "black",
        spacing: int = 4,
        halign: HAlignType = "center",
        valign: VAlignType = "center",
        lines_align: HAlignType = "left",
        stroke_ratio: float = 0,
        stroke_fill: Optional[ColorType] = None,
        font_fallback: bool = True,
        fontname: str = "",
        fallback_fonts: List[str] = [],
    ) -> "BuildImage":
        """
        在图片上指定区域画文字

        :参数:
          * ``xy``: 文字区域，顺序依次为 左，上，右，下
          * ``text``: 文字，支持多行
          * ``max_fontsize``: 允许的最大字体大小
          * ``min_fontsize``: 允许的最小字体大小
          * ``allow_wrap``: 是否允许折行
          * ``style``: 字体样式，默认为 "normal"
          * ``weight``: 字体粗细，默认为 "normal"
          * ``fill``: 文字颜色
          * ``spacing``: 多行文字间距
          * ``halign``: 横向对齐方式，默认为居中
          * ``valign``: 纵向对齐方式，默认为居中
          * ``lines_align``: 多行文字对齐方式，默认为靠左
          * ``stroke_ratio``: 文字描边的比例，即 描边宽度 / 字体大小
          * ``stroke_fill``: 描边颜色
          * ``font_fallback``: 是否使用后备字体，默认为 `True`
          * ``fontname``: 指定首选字体
          * ``fallback_fonts``: 指定备选字体
        """

        left = xy[0]
        top = xy[1]
        width = xy[2] - xy[0]
        height = xy[3] - xy[1]
        fontsize = max_fontsize
        while True:
            text2img = Text2Image.from_text(
                text,
                fontsize,
                style,
                weight,
                fill,
                spacing,
                lines_align,
                int(fontsize * stroke_ratio),
                stroke_fill,
                font_fallback,
                fontname,
                fallback_fonts,
            )
            text_w = text2img.width
            text_h = text2img.height
            if text_w > width and allow_wrap:
                text2img.wrap(width)
                text_w = text2img.width
                text_h = text2img.height
            if text_w > width or text_h > height:
                fontsize -= 1
                if fontsize < min_fontsize:
                    raise ValueError("在指定的区域和字体大小范围内画不下这段文字")
            else:
                x = left  # "left"
                if halign == "center":
                    x += (width - text_w) / 2
                elif halign == "right":
                    x += width - text_w

                y = top  # "top"
                if valign == "center":
                    y += (height - text_h) / 2
                elif valign == "bottom":
                    y += height - text_h

                self.paste(text2img.to_image(), (int(x), int(y)), alpha=True)
                return self

    def draw_bbcode_text(
        self,
        xy: XYType,
        text: str,
        max_fontsize: int = 30,
        min_fontsize: int = 12,
        allow_wrap: bool = False,
        fill: ColorType = "black",
        spacing: int = 4,
        halign: HAlignType = "center",
        valign: VAlignType = "center",
        lines_align: HAlignType = "left",
        stroke_ratio: float = 0,
        stroke_fill: Optional[ColorType] = None,
        font_fallback: bool = True,
        fontname: str = "",
        fallback_fonts: List[str] = [],
    ) -> "BuildImage":
        """
        在图片上指定区域画文字

        :参数:
          * ``xy``: 文字区域，顺序依次为 左，上，右，下
          * ``text``: 文字，支持多行
          * ``max_fontsize``: 允许的最大字体大小
          * ``min_fontsize``: 允许的最小字体大小
          * ``allow_wrap``: 是否允许折行
          * ``fill``: 文字颜色
          * ``spacing``: 多行文字间距
          * ``halign``: 横向对齐方式，默认为居中
          * ``valign``: 纵向对齐方式，默认为居中
          * ``lines_align``: 多行文字对齐方式，默认为靠左
          * ``stroke_ratio``: 文字描边的比例，即 描边宽度 / 字体大小
          * ``stroke_fill``: 描边颜色
          * ``font_fallback``: 是否使用后备字体，默认为 `True`
          * ``fontname``: 指定首选字体
          * ``fallback_fonts``: 指定备选字体
        """

        left = xy[0]
        top = xy[1]
        width = xy[2] - xy[0]
        height = xy[3] - xy[1]
        fontsize = max_fontsize
        while True:
            text2img = Text2Image.from_bbcode_text(
                text,
                fontsize,
                fill,
                spacing,
                lines_align,
                stroke_ratio,
                stroke_fill,
                font_fallback,
                fontname,
                fallback_fonts,
            )
            text_w = text2img.width
            text_h = text2img.height
            if text_w > width and allow_wrap:
                text2img.wrap(width)
                text_w = text2img.width
                text_h = text2img.height
            if text_w > width or text_h > height:
                fontsize -= 1
                if fontsize < min_fontsize:
                    raise ValueError("在指定的区域和字体大小范围内画不下这段文字")
            else:
                x = left  # "left"
                if halign == "center":
                    x += (width - text_w) / 2
                elif halign == "right":
                    x += width - text_w

                y = top  # "top"
                if valign == "center":
                    y += (height - text_h) / 2
                elif valign == "bottom":
                    y += height - text_h

                self.paste(text2img.to_image(), (int(x), int(y)), alpha=True)
                return self

    def save(self, format: str, **params) -> BytesIO:
        output = BytesIO()
        self.image.save(output, format, **params)
        return output

    def save_jpg(self, bg_color: ColorType = "white") -> BytesIO:
        """
        保存图片为 jpg 格式

        :参数:
          * ``bg_color``: 由 png 转为 jpg 时的背景颜色，默认为白色
        """
        if self.mode == "RGBA":
            img = self.new("RGBA", self.size, bg_color)
            img.paste(self.image, alpha=True)
        else:
            img = self
        return img.convert("RGB").save("jpeg")

    def save_png(self) -> BytesIO:
        """保存图片为 png 格式"""
        return self.convert("RGBA").save("png")



from typing import Tuple, Union, Literal


ModeType = Literal[
    "1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr"
]
ColorType = Union[str, Tuple[int, int, int], Tuple[int, int, int, int]]
PosTypeFloat = Tuple[float, float]
PosTypeInt = Tuple[int, int]
XYType = Tuple[float, float, float, float]
BoxType = Tuple[int, int, int, int]
PointsTYpe = Tuple[PosTypeFloat, PosTypeFloat, PosTypeFloat, PosTypeFloat]
DistortType = Tuple[float, float, float, float]
SizeType = Tuple[int, int]
HAlignType = Literal["left", "right", "center"]
VAlignType = Literal["top", "bottom", "center"]
OrientType = Literal["horizontal", "vertical"]
DirectionType = Literal[
    "center",
    "north",
    "south",
    "west",
    "east",
    "northwest",
    "northeast",
    "southwest",
    "southeast",
]
ResampleType = Literal[0, 1, 2, 3, 4, 5]
TransposeType = Literal[0, 1, 2, 3, 4, 5, 6]

FontStyle = Literal["normal", "italic", "oblique"]
FontWeight = Literal["ultralight", "light", "normal", "bold", "ultrabold", "heavy"]



import anyio
import httpx
import shutil
import traceback
from pathlib import Path
from PIL import ImageFont
from functools import lru_cache
from fontTools.ttLib import TTFont
from collections import namedtuple
from PIL.ImageFont import FreeTypeFont
from matplotlib.ft2font import FT2Font
from typing import List, Union, Optional, Set, Iterator
from matplotlib.font_manager import FontManager, FontProperties

from nonebot import get_driver
from nonebot.log import logger

from .config import Config
from .types import *

imageutils_config = Config.parse_obj(get_driver().config.dict())

FONT_PATH = imageutils_config.custom_font_path or Path("data/fonts")
FONT_PATH.mkdir(parents=True, exist_ok=True)

font_manager = FontManager()


def local_fonts() -> Iterator[str]:
    for f in FONT_PATH.iterdir():
        if f.is_file() and f.suffix in [".otf", ".ttf", ".ttc", ".afm"]:
            yield f.name


def add_font_to_manager(path: Union[str, Path]):
    try:
        if isinstance(path, Path):
            path = str(path.resolve())
        font_manager.addfont(path)
    except OSError as exc:
        logger.warning(f"Failed to open font file {path}: {exc}")
    except Exception as exc:
        logger.warning(f"Failed to extract font properties from {path}: {exc}")


for fontname in local_fonts():
    add_font_to_manager(FONT_PATH / fontname)


class Font:
    def __init__(self, family: str, fontpath: Path, valid_size: Optional[int] = None):
        self.family = family
        """字体族名字"""
        self.path = fontpath.resolve()
        """字体文件路径"""
        self.valid_size = valid_size
        """某些字体不支持缩放，只能以特定的大小加载"""
        self._glyph_table: Set[int] = set()
        for table in TTFont(self.path, fontNumber=0)["cmap"].tables:  # type: ignore
            for key in table.cmap.keys():
                self._glyph_table.add(key)

    @classmethod
    @lru_cache()
    def find(
        cls,
        family: str,
        style: FontStyle = "normal",
        weight: FontWeight = "normal",
        fallback_to_default: bool = True,
    ) -> "Font":
        """查找插件路径和系统路径下的字体"""
        font = cls.find_special_font(family)
        if font:
            return font
        font = cls.find_local_font(family)
        if font:
            return font
        font = cls.find_pil_font(family)
        if font:
            return font
        filepath = font_manager.findfont(
            FontProperties(family, style=style, weight=weight),  # type: ignore
            fallback_to_default=fallback_to_default,
        )
        font = FT2Font(filepath)
        return cls(font.family_name, Path(font.fname))

    @classmethod
    def find_local_font(cls, name: str) -> Optional["Font"]:
        """查找插件路径下的字体"""
        for fontname in local_fonts():
            if name == fontname or name == fontname.split(".")[0]:
                fontpath = FONT_PATH / fontname
                return cls(fontname, fontpath)

    @classmethod
    def find_pil_font(cls, name: str) -> Optional["Font"]:
        """通过 PIL ImageFont 查找系统字体"""
        try:
            font = ImageFont.truetype(name, 20)
            fontpath = Path(str(font.path))
            return cls(name, fontpath)
        except OSError:
            pass

    @classmethod
    def find_special_font(cls, family: str) -> Optional["Font"]:
        """查找特殊字体，主要是不可缩放的emoji字体"""

        SpecialFont = namedtuple("SpecialFont", ["family", "fontname", "valid_size"])
        SPECIAL_FONTS = {
            "Apple Color Emoji": SpecialFont(
                "Apple Color Emoji", "AppleColorEmoji.ttf", 137
            ),
            "Noto Color Emoji": SpecialFont(
                "Noto Color Emoji", "NotoColorEmoji.ttf", 109
            ),
        }

        if family in SPECIAL_FONTS:
            prop = SPECIAL_FONTS[family]
            fontname = prop.fontname
            valid_size = prop.valid_size
            fontpath = None
            if fontname in local_fonts():
                fontpath = FONT_PATH / fontname
            else:
                try:
                    font = ImageFont.truetype(fontname, valid_size)
                    fontpath = Path(str(font.path))
                except OSError:
                    pass
            if fontpath:
                return cls(family, fontpath, valid_size)

    @lru_cache()
    def load_font(self, fontsize: int) -> FreeTypeFont:
        """以指定大小加载字体"""
        return ImageFont.truetype(str(self.path), fontsize, encoding="utf-8")

    @lru_cache()
    def has_char(self, char: str) -> bool:
        """检查字体是否支持某个字符"""
        return ord(char) in self._glyph_table


default_fallback_fonts = imageutils_config.default_fallback_fonts


def get_proper_font(
    char: str,
    style: FontStyle = "normal",
    weight: FontWeight = "normal",
    fontname: Optional[str] = None,
    fallback_fonts: List[str] = [],
) -> Font:
    """
    获取合适的字体，将依次检查备选字体是否支持想要的字符

    :参数:
        * ``char``: 字符
        * ``style``: 字体样式，默认为 "normal"
        * ``weight``: 字体粗细，默认为 "normal"
        * ``fontname``: 可选，指定首选字体
        * ``fallback_fonts``: 可选，指定备选字体
    """
    fallback_fonts = fallback_fonts or default_fallback_fonts.copy()
    if fontname:
        fallback_fonts.insert(0, fontname)

    for family in fallback_fonts:
        try:
            font = Font.find(family, style, weight, fallback_to_default=False)
        except ValueError as e:
            logger.info(str(e))
            try:
                default_fallback_fonts.remove(family)
            except:
                pass
            continue
        if font.has_char(char):
            return font

    logger.warning(f"在当前字体列表中找不到可以显示字符“{char}”的字体")
    return Font.find("serif", style, weight)


async def add_font(fontname: str, source: Union[str, Path]):
    """通过字体文件路径或下载链接添加字体到插件路径"""
    fontpath = FONT_PATH / fontname
    if fontpath.exists():
        return
    try:
        if isinstance(source, Path):
            if source.is_file():
                shutil.copyfile(source, fontpath)
        else:
            await download_font(source, fontpath)
        add_font_to_manager(fontpath)
    except:
        logger.warning(
            f"Add font {fontname} from {source} failed\n{traceback.format_exc()}"
        )
        fontpath.unlink(missing_ok=True)


async def download_font(url: str, fontpath: Path):
    """下载字体到插件路径"""
    async with httpx.AsyncClient() as client:
        async with client.stream("GET", url) as resp:
            logger.info(f"Begin to download font from {url}")
            async with await anyio.open_file(fontpath, "wb") as file:
                async for chunk in resp.aiter_bytes():
                    await file.write(chunk)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_imageutils']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0.0,<10.0.0',
 'bbcode>=1.1.0,<2.0.0',
 'fonttools>=4.0.0,<5.0.0',
 'httpx>=0.19.0',
 'matplotlib>=3.0.0,<4.0.0',
 'nonebot2>=2.0.0-beta.1,<3.0.0',
 'numpy>=1.20.0,<2.0.0',
 'opencv-python-headless>=4.0.0,<5.0.0']

setup_kwargs = {
    'name': 'nonebot-plugin-imageutils',
    'version': '0.1.11',
    'description': 'Nonebot2 PIL工具插件',
    'long_description': '## nonebot-plugin-imageutils\n\n\n### 功能\n\n- 提供 `BuildImage` 类，方便图片尺寸修改、添加文字等操作\n- 提供 `Text2Image` 类，方便实现文字转图，支持少量 `BBCode` 标签\n- 文字支持多种字体切换，能够支持 `emoji`\n- 添加文字自动调节字体大小\n\n\n### 安装\n\n- 使用 nb-cli\n\n```\nnb plugin install nonebot_plugin_imageutils\n```\n\n- 使用 pip\n\n```\npip install nonebot_plugin_imageutils\n```\n\n\n### 配置字体\n\n本插件选择了一些不同系统上的字体，以支持更多的字符\n\n> 对于 `Ubuntu` 系统，建议安装 `fonts-noto` 软件包 以支持中文字体和 emoji\n>\n> 并将简体中文设置为默认语言：（否则会有部分中文显示为异体（日文）字形，详见 [ArchWiki](https://wiki.archlinux.org/title/Localization_(%E7%AE%80%E4%BD%93%E4%B8%AD%E6%96%87)/Simplified_Chinese_(%E7%AE%80%E4%BD%93%E4%B8%AD%E6%96%87)#%E4%BF%AE%E6%AD%A3%E7%AE%80%E4%BD%93%E4%B8%AD%E6%96%87%E6%98%BE%E7%A4%BA%E4%B8%BA%E5%BC%82%E4%BD%93%EF%BC%88%E6%97%A5%E6%96%87%EF%BC%89%E5%AD%97%E5%BD%A2)）\n> ```bash\n> sudo apt install fonts-noto\n> sudo locale-gen zh_CN zh_CN.UTF-8\n> sudo update-locale LC_ALL=zh_CN.UTF-8 LANG=zh_CN.UTF-8\n> fc-cache -fv\n> ```\n\n默认备选字体列表如下：\n```\n"Arial", "Tahoma", "Helvetica Neue", "Segoe UI", "PingFang SC", "Hiragino Sans GB", "Microsoft YaHei", "Source Han Sans SC", "Noto Sans SC", "Noto Sans CJK JP", "WenQuanYi Micro Hei", "Apple Color Emoji", "Noto Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol"\n```\n\n可在 `.env.*` 文件中添加 `default_fallback_fonts` 变量 来自定义备选字体\n\n字体文件需要在系统目录下，或放置于自定义字体路径中\n\n自定义字体路径默认为机器人运行目录下的 `data/fonts/` 文件夹，\n\n可在 `.env.*` 文件中添加 `custom_font_path` 变量 自定义字体路径\n\n其他插件可以通过 `nonebot_plugin_imageutils/fonts.py` 中的 `add_font` 函数往字体文件夹中添加字体\n\n\n### 使用示例\n\n\n- `BuildImage`\n\n```python\nfrom nonebot_plugin_imageutils import BuildImage\n\n# output: BytesIO\noutput = BuildImage.new((300, 300)).circle().draw_text((30, 30, 270, 270), "测试ymddl😂").save_jpg()\n```\n\n![1.jpg](https://s2.loli.net/2022/05/19/gFdpwWPCzreb2X6.jpg)\n\n\n- `Text2Image`\n\n```python\nfrom nonebot_plugin_imageutils import Text2Image\n\n# img: PIL.Image.Image\nimg = Text2Image.from_text("@mnixry 🤗", 50).to_image()\n\n# 以上结果为 PIL 的 Image 格式，若要直接 MessageSegment 发送，可以转为 BytesIO\noutput = BytesIO()\nimg.save(output, format="png")\nawait matcher.send(MessageSegment.image(output))\n```\n\n![2.png](https://s2.loli.net/2022/05/19/14EXViZQwcGUW5I.png)\n\n\n- 使用 `BBCode`\n\n```python\nfrom nonebot_plugin_imageutils import text2image\n\n# img: PIL.Image.Image\nimg = text2image("N[size=40][color=red]o[/color][/size]neBo[size=30][color=blue]T[/color][/size]\\n[align=center]太强啦[/align]")\n\n# 以上结果为 PIL 的 Image 格式，若要直接 MessageSegment 发送，可以转为 BytesIO\noutput = BytesIO()\nimg.save(output, format="png")\nawait matcher.send(MessageSegment.image(output))\n```\n\n![3.png](https://s2.loli.net/2022/05/19/VZAXsKB2x65q7rl.png)\n\n目前支持的 `BBCode` 标签：\n- `[align=left|right|center][/align]`: 文字对齐方式\n- `[color=#66CCFF|red|black][/color]`: 字体颜色\n- `[stroke=#66CCFF|red|black][/stroke]`: 描边颜色\n- `[font=msyh.ttc][/font]`: 文字字体\n- `[size=30][/size]`: 文字大小\n- `[b][/b]`: 文字加粗\n\n\n### 特别感谢\n\n- [HibiKier/zhenxun_bot](https://github.com/HibiKier/zhenxun_bot) 基于 Nonebot2 和 go-cqhttp 开发，以 postgresql 作为数据库，非常可爱的绪山真寻bot\n',
    'author': 'meetwq',
    'author_email': 'meetwq@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/noneplugin/nonebot-plugin-imageutils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
from bbcode import Parser
from PIL import Image, ImageDraw
from PIL.Image import Image as IMG
from PIL.ImageColor import colormap
from typing import List, Optional, Iterator

from .types import *
from .fonts import Font, get_proper_font


class Char:
    def __init__(
        self,
        char: str,
        font: Font,
        fontsize: int = 16,
        fill: ColorType = "black",
        stroke_width: int = 0,
        stroke_fill: Optional[ColorType] = None,
    ):
        self.char = char
        self.font = font
        self.fontsize = fontsize
        self.fill = fill
        self.stroke_width = stroke_width
        self.stroke_fill = stroke_fill

        if self.font.valid_size:
            self.stroke_width = 0
            self.pilfont = self.font.load_font(self.font.valid_size)
        else:
            self.pilfont = self.font.load_font(fontsize)

        self.ascent, self.descent = self.pilfont.getmetrics()
        self.width, self.height = self.pilfont.getsize(
            self.char, stroke_width=self.stroke_width
        )

        if self.font.valid_size:
            ratio = fontsize / self.font.valid_size
            self.ascent = round(self.ascent * ratio)
            self.descent = round(self.descent * ratio)
            self.width = round(self.width * ratio)
            self.height = round(self.height * ratio)

    def draw_on(self, img: IMG, pos: PosTypeInt):
        if self.font.valid_size:
            ratio = self.font.valid_size / self.fontsize
            new_img = Image.new(
                "RGBA", (int(self.width * ratio), int(self.height * ratio))
            )
            draw = ImageDraw.Draw(new_img)
            draw.text(
                (0, 0),
                self.char,
                font=self.pilfont,
                fill=self.fill,
                embedded_color=True,
            )
            new_img = new_img.resize(
                (int(self.width), int(self.height)), resample=Image.ANTIALIAS
            )
            img.paste(new_img, pos, mask=new_img)
        else:
            draw = ImageDraw.Draw(img)
            draw.text(
                pos,
                self.char,
                font=self.pilfont,
                fill=self.fill,
                stroke_width=self.stroke_width,
                stroke_fill=self.stroke_fill,
                embedded_color=True,
            )


class Line:
    def __init__(
        self, chars: List[Char], align: HAlignType = "left", fontsize: int = 16
    ):
        self.chars: List[Char] = chars
        self.align: HAlignType = align
        self.fontsize = fontsize

    @property
    def width(self) -> int:
        if not self.chars:
            return 0
        return (
            sum([char.width - char.stroke_width * 2 for char in self.chars])
            + self.chars[0].stroke_width
            + self.chars[-1].stroke_width
        )

    @property
    def height(self) -> int:
        if not self.chars:
            return Char("A", get_proper_font("A"), fontsize=self.fontsize).height
        return max([char.ascent + char.stroke_width for char in self.chars]) + max(
            [char.descent + char.stroke_width for char in self.chars]
        )

    @property
    def ascent(self) -> int:
        if not self.chars:
            return Char("A", get_proper_font("A"), fontsize=self.fontsize).ascent
        return max([char.ascent for char in self.chars])

    @property
    def descent(self) -> int:
        if not self.chars:
            return Char("A", get_proper_font("A"), fontsize=self.fontsize).descent
        return max([char.descent for char in self.chars])

    def wrap(self, width: float) -> Iterator["Line"]:
        last_idx = 0
        for idx in range(len(self.chars)):
            if Line(self.chars[last_idx : idx + 1]).width > width:
                yield Line(self.chars[last_idx:idx], self.align)
                last_idx = idx
        yield Line(self.chars[last_idx:], self.align)


class Text2Image:
    def __init__(self, lines: List[Line], spacing: int = 4):
        self.lines = lines
        self.spacing = spacing

    @classmethod
    def from_text(
        cls,
        text: str,
        fontsize: int,
        style: FontStyle = "normal",
        weight: FontWeight = "normal",
        fill: ColorType = "black",
        spacing: int = 4,
        align: HAlignType = "left",
        stroke_width: int = 0,
        stroke_fill: Optional[ColorType] = None,
        font_fallback: bool = True,
        fontname: str = "",
        fallback_fonts: List[str] = [],
    ) -> "Text2Image":
        """
        从文本构建 `Text2Image` 对象

        :参数:
          * ``text``: 文本
          * ``fontsize``: 字体大小
          * ``style``: 字体样式，默认为 "normal"
          * ``weight``: 字体粗细，默认为 "normal"
          * ``fill``: 文字颜色
          * ``spacing``: 多行文字间距
          * ``align``: 多行文字对齐方式，默认为靠左
          * ``stroke_width``: 文字描边宽度
          * ``stroke_fill``: 描边颜色
          * ``font_fallback``: 是否使用后备字体，默认为 `True`
          * ``fontname``: 指定首选字体
          * ``fallback_fonts``: 指定备选字体
        """

        font = None
        if not font_fallback:
            if not fontname:
                raise ValueError("`font_fallback` 为 `False` 时必须指定 `fontname`")
            font = Font.find(fontname, fallback_to_default=False)

        lines: List[Line] = []
        chars: List[Char] = []

        text = "\n".join(text.splitlines())
        for char in text:
            if char == "\n":
                lines.append(Line(chars, align))
                chars = []
                continue
            if font_fallback:
                font = get_proper_font(char, style, weight, fontname, fallback_fonts)
            else:
                assert font
            chars.append(Char(char, font, fontsize, fill, stroke_width, stroke_fill))
        if chars:
            lines.append(Line(chars, align, fontsize))
        return cls(lines, spacing)

    @classmethod
    def from_bbcode_text(
        cls,
        text: str,
        fontsize: int = 30,
        fill: ColorType = "black",
        spacing: int = 6,
        align: HAlignType = "left",
        stroke_ratio: float = 0,
        stroke_fill: Optional[ColorType] = None,
        font_fallback: bool = True,
        fontname: str = "",
        fallback_fonts: List[str] = [],
    ) -> "Text2Image":
        """
        从含有 `BBCode` 的文本构建 `Text2Image` 对象

        目前支持的 `BBCode` 标签：
          * ``[align=left|right|center][/align]``: 文字对齐方式
          * ``[color=#66CCFF|red|black][/color]``: 字体颜色
          * ``[stroke=#66CCFF|red|black][/stroke]``: 描边颜色
          * ``[font=msyh.ttc][/font]``: 文字字体，需填写完整字体文件名
          * ``[size=30][/size]``: 文字大小
          * ``[b][/b]``: 文字加粗

        :参数:
          * ``text``: 文本
          * ``fontsize``: 字体大小，默认为 30
          * ``fill``: 文字颜色，默认为 `black`
          * ``spacing``: 多行文字间距
          * ``align``: 多行文字对齐方式，默认为靠左
          * ``stroke_ratio``: 文字描边的比例，即 描边宽度 / 字体大小
          * ``stroke_fill``: 描边颜色
          * ``font_fallback``: 是否使用后备字体，默认为 `True`
          * ``fontname``: 指定首选字体
          * ``fallback_fonts``: 指定备选字体
        """

        font = None
        if not font_fallback:
            if not fontname:
                raise ValueError("`font_fallback` 为 `False` 时必须指定 `fontname`")
            font = Font.find(fontname, fallback_to_default=False)

        lines: List[Line] = []
        chars: List[Char] = []

        def new_line():
            nonlocal lines
            nonlocal chars
            lines.append(Line(chars, last_align, fontsize))
            chars = []

        align_stack = []
        color_stack = []
        stroke_stack = []
        font_stack = []
        size_stack = []
        bold_stack = []
        last_align: HAlignType = align

        align_pattern = r"left|right|center"
        colors = "|".join(colormap.keys())
        color_pattern = rf"#[a-fA-F0-9]{{6}}|{colors}"
        stroke_pattern = color_pattern
        font_pattern = r"\S+\.ttf|\S+\.ttc|\S+\.otf|\S+\.fnt"
        size_pattern = r"\d+"

        parser = Parser()
        parser.recognized_tags = {}
        parser.add_formatter("align", None)
        parser.add_formatter("color", None)
        parser.add_formatter("stroke", None)
        parser.add_formatter("font", None)
        parser.add_formatter("size", None)
        parser.add_formatter("b", None)

        text = "\n".join(text.splitlines())
        tokens = parser.tokenize(text)
        for token_type, tag_name, tag_opts, token_text in tokens:
            if token_type == 1:
                if tag_name == "align":
                    if re.fullmatch(align_pattern, tag_opts["align"]):
                        align_stack.append(tag_opts["align"])
                elif tag_name == "color":
                    if re.fullmatch(color_pattern, tag_opts["color"]):
                        color_stack.append(tag_opts["color"])
                elif tag_name == "stroke":
                    if re.fullmatch(stroke_pattern, tag_opts["stroke"]):
                        stroke_stack.append(tag_opts["stroke"])
                elif tag_name == "font":
                    if re.fullmatch(font_pattern, tag_opts["font"]):
                        font_stack.append(tag_opts["font"])
                elif tag_name == "size":
                    if re.fullmatch(size_pattern, tag_opts["size"]):
                        size_stack.append(tag_opts["size"])
                elif tag_name == "b":
                    bold_stack.append(True)
            elif token_type == 2:
                if tag_name == "align":
                    if align_stack:
                        align_stack.pop()
                elif tag_name == "color":
                    if color_stack:
                        color_stack.pop()
                elif tag_name == "stroke":
                    if stroke_stack:
                        stroke_stack.pop()
                elif tag_name == "font":
                    if font_stack:
                        font_stack.pop()
                elif tag_name == "size":
                    if size_stack:
                        size_stack.pop()
                elif tag_name == "b":
                    if bold_stack:
                        bold_stack.pop()
            elif token_type == 3:
                new_line()
            elif token_type == 4:
                char_align = align_stack[-1] if align_stack else align
                char_color = color_stack[-1] if color_stack else fill
                char_stroke = stroke_stack[-1] if stroke_stack else stroke_fill
                char_font = font_stack[-1] if font_stack else fontname
                char_size = size_stack[-1] if size_stack else fontsize
                char_bold = bold_stack[-1] if bold_stack else False

                if char_align != last_align:
                    if chars:
                        new_line()
                    last_align = char_align
                for char in token_text:
                    if font_fallback:
                        font = get_proper_font(
                            char,
                            weight="bold" if char_bold else "normal",
                            fontname=char_font,
                            fallback_fonts=fallback_fonts,
                        )
                    else:
                        assert font
                    chars.append(
                        Char(
                            char,
                            font,
                            int(char_size),
                            char_color,
                            int(char_size * stroke_ratio),
                            char_stroke,
                        )
                    )

        if chars:
            new_line()

        return cls(lines, spacing)

    @property
    def width(self) -> int:
        if not self.lines:
            return 0
        return max([line.width for line in self.lines])

    @property
    def height(self) -> int:
        if not self.lines:
            return 0
        return (
            sum([line.ascent for line in self.lines])
            + self.lines[-1].descent
            + self.spacing * (len(self.lines) - 1)
            + max([char.stroke_width for char in self.lines[0].chars])
            + max([char.stroke_width for char in self.lines[-1].chars])
        )

    def wrap(self, width: float) -> "Text2Image":
        new_lines: List[Line] = []
        for line in self.lines:
            new_lines.extend(line.wrap(width))
        self.lines = new_lines
        return self

    def to_image(
        self,
        bg_color: Optional[ColorType] = None,
        padding: Union[SizeType, BoxType] = (0, 0),
    ) -> IMG:
        if len(padding) == 4:
            padding_left, padding_top, padding_right, padding_bottom = padding
        else:
            padding_left = padding_right = padding[0]
            padding_top = padding_bottom = padding[1]

        img = Image.new(
            "RGBA",
            (
                self.width + padding_left + padding_right,
                self.height + padding_top + padding_bottom,
            ),
            bg_color,  # type: ignore
        )

        top = padding_top
        for line in self.lines:
            left = padding_left
            if line.align == "center":
                left += (self.width - line.width) / 2
            elif line.align == "right":
                left += self.width - line.width

            x = left
            if line.chars:
                x += line.chars[0].stroke_width
            for char in line.chars:
                y = top + line.ascent - char.ascent
                char.draw_on(img, (int(x), int(y)))
                x += char.width - char.stroke_width * 2
            top += line.ascent + self.spacing

        return img


def text2image(
    text: str,
    bg_color: ColorType = "white",
    padding: Union[SizeType, BoxType] = (10, 10),
    max_width: Optional[int] = None,
    font_fallback: bool = True,
    **kwargs,
) -> IMG:
    """
    文字转图片，支持少量 `BBCode` 标签，具体见 `Text2Image` 类的 `from_bbcode_text` 函数

    :参数:
        * ``text``: 文本
        * ``bg_color``: 图片背景颜色
        * ``padding``: 图片边距
        * ``max_width``: 图片最大宽度，不设置则不限宽度
        * ``font_fallback``: 是否使用后备字体，默认为 `True`
    """
    text2img = Text2Image.from_bbcode_text(text, font_fallback=font_fallback, **kwargs)
    if max_width:
        text2img.wrap(max_width)
    return text2img.to_image(bg_color, padding)



from pathlib import Path
from typing import List, Optional
from pydantic import BaseModel, Extra


class Config(BaseModel, extra=Extra.ignore):
    custom_font_path: Optional[Path] = None
    default_fallback_fonts: List[str] = [
        "Arial",
        "Tahoma",
        "Helvetica Neue",
        "Segoe UI",
        "PingFang SC",
        "Hiragino Sans GB",
        "Microsoft YaHei",
        "Source Han Sans SC",
        "Noto Sans SC",
        "Noto Sans CJK JP",
        "WenQuanYi Micro Hei",
        "Apple Color Emoji",
        "Noto Color Emoji",
        "Segoe UI Emoji",
        "Segoe UI Symbol",
    ]
